/*----------------------------------------------------------------------*/
/* Ev_Board.c                                                           */
/*                                                                      */
/* Ev.Board: Raspberry pi PICO                                          */
/* Console: N81 115200bps                                               */
/* MMC/SDC: SPI mode                      K.S. (c)2021                  */
/*----------------------------------------------------------------------*/

#include "ff.h"
#include "Ev_Board_conf.h"
#include <ctype.h>
#include <string.h>
#include <unistd.h>

extern void disk_timerproc (void);
extern void tte(char *fnm);

volatile UINT Timer = 0;		/* Performance counter (1kHz increment) */
static   unsigned long Sec_Timer = 0;

/*---------------------------------------------*/
/* 1kHz timer process                          */
/*---------------------------------------------*/

void Disk_Timer_Handler (void)
{
	static UINT led;

    EVBOARD_CLEAR_INTFLAG();

	Timer++;	/* Increment performance counter */
        if ( Timer >= 1000 ) {
             Sec_Timer++;  // 1 Sec Timer
             Timer = 0;
        }

	led++;		/* LED blinking timer */
	if ((led & 0x40) && (led & 0x680) == 0) {
		LED_ON();
	} else {
		LED_OFF();
	}

	disk_timerproc();	/* Disk timer process (mmc.c) */
}

void Ev_Board_Timer_init(irq_handler_t handler)
{
   gpio_set_function(PICO_DEFAULT_LED_PIN, GPIO_FUNC_PWM);

    uint slice_num = pwm_gpio_to_slice_num(PICO_DEFAULT_LED_PIN);

    pwm_clear_irq(slice_num);
    pwm_set_irq_enabled(slice_num, true);
    irq_set_exclusive_handler(PWM_IRQ_WRAP, handler);
    irq_set_enabled(PWM_IRQ_WRAP, true);
    pwm_set_wrap(slice_num, FPWM);
    pwm_set_clkdiv(slice_num, 100.f);
    pwm_set_gpio_level(PICO_DEFAULT_LED_PIN, (FPWM+1)/100); // 1% duty
    // Set the PWM running
    pwm_set_enabled(slice_num, true);
}

/*-----------------------------------------------------------------------*/
/* HARD Fault Function                                                   */
/*-----------------------------------------------------------------------*/
void Ev_Board_Hardfault_Handler(void)
{
   unsigned long x;

   PWM4_CNTL(0);		// stop PWM
   GPIO_CTL25(5);  	    // set GPIO25 SIO-MODE
   SIO_GPIO_OE(LED_PIN);	// set GPIO25 OUTPUT PIN
   while(1) {
      SIO_GPIO_OUT_XOR(LED_PIN);  // blink GPIO25 LED
      for ( x = 0 ; x < 5050050; x++ ) asm volatile("nop \n");
   }
}

/*-----------------------------------------------------------------------*/
/* Initialize MMC controls                                               */
/*-----------------------------------------------------------------------*/
void spi_port_ini(void)
{
    gpio_init(CS_PIN);
    gpio_set_dir(CS_PIN, GPIO_OUT);
    gpio_put(CS_PIN, 1);
    spi_init(SPI_PORT, 500 * 1000);
    spi_set_format(SPI_PORT, 8, SPI_CPOL_0, SPI_CPHA_0, SPI_MSB_FIRST); // SPI-mode0
    gpio_set_function(MISO_PIN, GPIO_FUNC_SPI);
    gpio_set_function(SCK_PIN, GPIO_FUNC_SPI);
    gpio_set_function(MOSI_PIN, GPIO_FUNC_SPI);
    gpio_set_function(CS_PIN, GPIO_FUNC_SPI);
    gpio_pull_up(MISO_PIN);
}

/*-----------------------------------------------------------------------*/
/* MMC controls                                                          */
/*-----------------------------------------------------------------------*/

/* Exchange a byte */
BYTE xchg_spi (
	BYTE dat	/* Data to send */
)
{
    BYTE r;
    while(!spi_is_writable(SPI_PORT) ) ;
    spi_write_read_blocking( SPI_PORT, &dat, &r, 1 );
    return r;
}


/* Receive multiple byte */
void rcvr_spi_multi (
	BYTE *buff,		/* Pointer to data buffer */
	UINT btr		/* Number of bytes to receive (16, 64 or 512) */
)
{

	while (btr) {
		*buff++ = xchg_spi(0xFF);
		btr--;
	}
}


#if FF_FS_READONLY == 0
/* Send multiple byte */
void xmit_spi_multi (
	const BYTE *buff,	/* Pointer to the data */
	UINT btx			/* Number of bytes to send (512) */
)
{

	while (btx) {
		xchg_spi(*buff++);
		btx--;
	}
}
#endif

/*-----------------------------------------------------------------------*/
/* UART controls                                                         */
/*-----------------------------------------------------------------------*/
void Ev_Board_uart_init(void)
{
    uart_init(UART_ID, BAUD_RATE);

    gpio_set_function(UART_TX_PIN, GPIO_FUNC_UART);
    gpio_set_function(UART_RX_PIN, GPIO_FUNC_UART);
    sleep_us(500);
    if( uart_is_readable (UART_ID) )    uart_getc(UART_ID);
    if( uart_is_readable (UART_ID) )    uart_getc(UART_ID);
}

/*-----------------------------------------------------------------------*/
/* RTC controls                                                          */
/*-----------------------------------------------------------------------*/
#if !FF_FS_NORTC && !FF_FS_READONLY
uint8_t get_wday(datetime_t *t)
{
    uint16_t y, m, d, wday;
    y = t->year;
    m = t->month;
    d = t->day;
    if ( m <= 2 ) {
        m += 12;
        y--;
    }
    wday = y + y/4 - y / 100 + y / 400 + ( 13 * m + 8) / 5 + d;
    wday %= 7;
    return (uint8_t)wday;
}
static datetime_t t_rtc = {
          .year  = 2021,
          .month = 03,
          .day   = 28,
          .dotw  = 0,
          .hour  = 0,
          .min   = 0,
          .sec   = 0
       };


int set_rtc(char *ctime)
{
    int y, m, d, h, min, s;

    if( sscanf(ctime,"%d/%d/%d %d:%d:%d", &y, &m, &d, &h, &min, &s) != 6 ) return 1;
    if ( y < 2021 || y > 2399 || m < 1 || m > 12 || h < 0 || 
          h > 23 || min < 0 || min > 59 || s < 0 || s > 59)  return 1;
    if( (m == 1 || m == 3 || m == 5 || m == 7 || m == 8|| m == 10 || m == 12) &&
         (d < 1 || d > 31 ) )   return 1;
    if( (m == 4 || m == 6 || m == 9 || m == 11) &&
         (d < 1 || d > 30 ) )   return 1;
    if( m == 2 ) {
        if ( (y % 4) == 0 ) {
            if ( d < 1 || d > 29 ) return 1;
        } else {
            if ( d < 1 || d > 28 ) return 1;
        }
    }
    rtc_init();

    t_rtc.year = y;
    t_rtc.month = m;
    t_rtc.day = d;
    t_rtc.hour = h;
    t_rtc.min = min;
    t_rtc.sec = s;

    t_rtc.dotw = get_wday(&t_rtc);
    rtc_set_datetime(&t_rtc);
    return 0;
}
DWORD get_rtc_date(void)
{
    rtc_get_datetime(&t_rtc);
    return t_rtc.year*16*256+t_rtc.month*256+t_rtc.day;
}
DWORD get_rtc_time(void)
{
    rtc_get_datetime(&t_rtc);
    return t_rtc.hour*256*256+t_rtc.min*256+t_rtc.sec;
}

DWORD get_rtime(void)
{
    return Sec_Timer; 
}
DWORD get_fattime (void)
{
    rtc_get_datetime(&t_rtc);
    return	  ((DWORD)(t_rtc.year - 1980) << 25)	/* Return fixed value 2012/4/1 0:0:0 */
			| ((DWORD)t_rtc.month << 21)
			| ((DWORD)t_rtc.day << 16)
			| ((DWORD)t_rtc.hour << 11)
			| ((DWORD)t_rtc.min << 5)
			| ((DWORD)t_rtc.sec >> 1);
}
#endif
/*-----------------------------------------------------------------------*/
/* File Open controls                                                          */
/*-----------------------------------------------------------------------*/

static char  F_ready_on;
static FATFS FatFs;
static struct file_obj file_object[5];

void fatfs_init(void)
{
    void *p;

    for ( int i = 0; i < 5; i++ ) {
        file_object[i].opn_f = 0;
        file_object[i].end_f = 0;
    }
    memset((void*)&FatFs,0,sizeof(FATFS));
    f_mount(&FatFs, "", 0);
    F_ready_on = 1;
}
void fatfs_Not_ready(void)
{
    F_ready_on = 0;
}
int fatfs_is_ready(void)
{
    return F_ready_on;
}
FRESULT F_Open (
	int file_no,		/* file object number */
	const TCHAR* path,	/* Pointer to the file name */
	const TCHAR* mode	/* Access mode and file open mode flags */
)
{
    FRESULT re;
    const char fmode[] = "r\000r+\000w\000w+\000a\000a+\000wx\000w+x\000\000";
    int i, m;
    char fml[4];
    for ( i = 0; i < 3 && *(mode+i) ; i++ ) {
        m = *(mode+i);
        fml[i] = tolower(m);
    }
    fml[i] = 0;
    for ( m = 0, i = 0; *(fmode+i); m++ ) {
        if(strcmp(fml, fmode+i) == 0) break;
        for( i++; *(fmode+i); i++ ) ;
        i++;
    }
    switch(m) {
    case 0:  // "r"
        m = FA_READ;
        break;
    case 1:  // "r+"
        m = FA_READ | FA_WRITE;
        break;
    case 2:  // "w"
        m = FA_CREATE_ALWAYS | FA_WRITE;
        break;
    case 3:  // "w+"
        m = FA_CREATE_ALWAYS | FA_WRITE | FA_READ;
        break;
    case 4:  // "a"
        m = FA_OPEN_APPEND | FA_WRITE;
        break;
    case 5:  // "a+"
        m = FA_OPEN_APPEND | FA_WRITE | FA_READ;
        break;
    case 6:  // "wx"
        m = FA_CREATE_NEW | FA_WRITE;
        break;
    case 7:  // "w+x"
        m = FA_CREATE_NEW | FA_WRITE | FA_READ;
        break;
    default:
        return FR_DISK_ERR;
    }
    if( file_object[file_no].opn_f != 0 ) {
        return FR_DISK_ERR;
    }
    if( FILENO_CHECK(file_no) ) return  FR_DISK_ERR;
    if( !fatfs_is_ready() )       return  FR_DISK_ERR;
    re = f_open(&file_object[file_no].fil, path, m);
    if ( !re ) {
        file_object[file_no].opn_f = 1;
        file_object[file_no].end_f = 0;
    }
    return re;
}
int F_write(int fn, unsigned char c)
{
    UINT    wsiz;

    if( FILENO_CHECK(fn) ) return 1;
    if( file_object[fn].opn_f != 0 ) {
        f_write(&file_object[fn].fil, &c, 1, &wsiz);            /* Write it to the destination file */
        if (wsiz != 1)  return 1; /* error or disk full */
    } else {
        return 1;
    }
    return 0;
}
int F_writeN(int fn, unsigned char *buf, UINT sz)
{
    UINT    wsiz;

    if( FILENO_CHECK(fn) ) return 1;
    if( file_object[fn].opn_f != 0 ) {
        f_write(&file_object[fn].fil, buf, sz, &wsiz);            /* Write it to the destination file */
        if (wsiz != sz)  return 1; /* error or disk full */
    } else {
        return 1;
    }
    return 0;
}

int F_puts(int fn, unsigned char *buf)
{
    UINT    dsiz, wsiz;
    dsiz = strlen(buf);
    if( FILENO_CHECK(fn) ) return 1;
    if( file_object[fn].opn_f != 0 ) {
        f_write(&file_object[fn].fil, buf, dsiz, &wsiz);            /* Write it to the destination file */
        if (wsiz != dsiz)  return 1; /* error or disk full */
        f_write(&file_object[fn].fil, "\r\n", 2, &wsiz);            /* Write it to the destination file */
        if (wsiz != 2)  return 1; /* error or disk full */
    } else {
        return 1;
    }
    return 0;
}
int F_gets( unsigned char *buf,int siz, int fn)
{
    UINT    rsiz, len;
    int     cf;
    len = 0;
    if (siz < 4)           return 1;
    siz -= 3;
    if( FILENO_CHECK(fn) ) return 1;
    if( file_object[fn].opn_f != 0 ) {
        while(siz--) {
          cf = 0;
          f_read(&file_object[fn].fil, &cf, 1, &rsiz);  /* Read a chunk of data from the source file */
          if (rsiz == 0)     break; /* error or eof */
          if ( cf == 0x0a )  break;
          len++;
          *buf++ = cf;
        }
        if( *(buf-1) == 0x0d ) { buf--; *buf = 0; }
        if( cf == 0x0a ) {
           *buf++ = 0x0d; *buf++ = 0x0a;
        }
        *buf = 0;
        if( rsiz == 0 && len == 0 )   return 1;
     } else {
        return 1;
    }
    return 0;
}

int F_read(int fn, unsigned char *c)
{
    UINT    rsiz;
    int     cf;

    if( FILENO_CHECK(fn) ) return 1;
    if( file_object[fn].opn_f != 0 ) {
        f_read(&file_object[fn].fil, &cf, 1, &rsiz);  /* Read a chunk of data from the source file */
        if (rsiz == 0) return 1; /* error or eof */
        *c = cf;
    } else {
        return 1;
    }
    return 0;
}
int F_copy( char *src, char *dst )
{
  FIL fsrc, fdst;      /* File objects */
  FRESULT fr;          /* FatFs function common result code */
  UINT br, bw;         /* File read/write count */
  BYTE buffer[4096];   /* File copy buffer */

  if( !fatfs_is_ready() ) {
    printf(" SD Card Not Ready ??\n");
    return  0;
  }
  fr = f_open(&fsrc, src, FA_READ);
  if (fr) {
    printf(" cp : Open %s Error? ", src);
    return 0;
  }
  /* Create destination file on the drive 0 */
  fr = f_open(&fdst, dst, FA_WRITE | FA_CREATE_ALWAYS);
  if (fr) {
    f_close(&fsrc);
    printf(" cp : Open %s Error? ", dst);
    return 0;
  }
  /* Copy source to destination */
  for (;;) {
    f_read(&fsrc, buffer, sizeof buffer, &br);  /* Read a chunk of data from the source file */
    if (br == 0) break; /* error or eof */
    f_write(&fdst, buffer, br, &bw);            /* Write it to the destination file */
    if (bw < br) {
      printf(" ERROR %s Disk Full !\n", dst);
      break; /* error or disk full */
    }
  }   /* Close open files */
  /* Close open files */
  f_close(&fsrc);
  f_close(&fdst);
  return 1;
}
static FRESULT F_type(char *fn)
{
    FRESULT rfr;         /* FatFs return code */
    char line[640]; /* Line buffer */
    int rsiz;

    if( !fatfs_is_ready() ) {
      printf(" SD Card Not Ready ??\n");
      return  0;
    }
    /* Open a text file */
    rfr = F_Open(3, fn, "r");
    if (rfr) return rfr;
    while (!F_gets(line, sizeof(line), 3)) {
        if( write(STDOUT_FILENO, line, strlen(line)) == -1 ) {
           printf("\r\n\r\nWrite Error STDOUT !! \r\n");
           break;
        }
        if( getchar_timeout_us(0) == 0x03 )  {
            printf("\r\n");
            break;
        }
    }
    /* Close the file */
    F_close(3);
    return FR_OK;
}
void F_close(int fn)
{
    if( !fatfs_is_ready() ) {
      printf(" SD Card Not Ready ??\n");
      return;
    }
    if( FILENO_CHECK(fn) ) return;
    if( file_object[fn].opn_f != 0 ) {
        f_close(&file_object[fn].fil);
        file_object[fn].opn_f = 0;
        file_object[fn].end_f = 0;
    }
}

/*-----------------------------------------------------------------------*/
/* File Operation                                                        */
/*-----------------------------------------------------------------------*/
#define C_MKDIR 0
#define C_RMDIR 1
#define C_TYPE 2
#define C_CP 3
#define C_RM 4
#define C_MOUNT 5
#define C_EDIT 6
#define C_NONE 7

static char * px_string(char *s, int *l)
{
  char d, *p;

  *l = -1;
  p = s;
  if(*p == '\'' || *p == '"' )
    d = *p++;
  else return s;
  *l = 0;
  while( *p && *p != d ) { *l += 1; p++; }
  if( *p && *p == d ) {
    return s+1;
  }
  *l = -1;
  return s;
}
void myputs(char *s)
{
    while( *s != 0x0a && *s != 0 ) { printf("%c", *s++);}
}

static char * skip_sp(char *s)
{
  while( *s == ' ' ) s++;
  return s;
}

static char * get_cmd(char *s, int *cmd_c)
{
  static char *cmdcmd = "mkdir\000rmdir\000type\000cp\000rm\000mount\000edit\000\000";

    int i, j, m;
    char *c;

    char fml[4];
    for ( j = 0, i = 0; *(cmdcmd+j) ; i++ ) {
        c = s;
        for ( ; *(cmdcmd+j) ; j++ ) {
            if( *(cmdcmd+j) != tolower(*c) )  break ;
            c++;
        }
        if( *(cmdcmd+j) == 0 && ( *c == 0 || *c == ' ' || *c == 0x0A) ) {
            if( *c == 0x0A ) s = c;
            else             s = *c ? c + 1: c;
            break;
        }
        for ( ; *(cmdcmd+j) ; j++ ) ;
	j++;
    }
    *cmd_c = i;
    return s;
}

int file_operate(unsigned char *cmd)
{
  char *nxt, *p1, *p2, *ps;
  int ex, ln1, ln2;

  ps = cmd;
  nxt = get_cmd(cmd, &ex);
  if( ex == C_NONE ) {
    printf(" ??Command Error!! :[");
    myputs(nxt);
    printf("] ");
    return 0;
  }
  nxt = skip_sp(nxt);
  p1 = px_string(nxt, &ln1);
  if( ln1 < 0 ) {
      if( *p1 == 0x0a ) {
          if( ex == C_MOUNT ) {
            if ( !fatfs_is_ready() ) {
               fatfs_init();
            }
            return (p1-ps);
          } else if ( ex == C_EDIT ) {
              ln1 = 0;
              nxt = p1;
              p1 = "";
          }
      } else nxt = p1;
  } else {
    nxt = p1 + ln1 + 1;
  }
  if ( ln1 >= 0 ) {
    *(p1+ln1) = 0;
  } else {
    printf(" ??Command File Name 1:[");
    myputs(nxt);
    printf("] ");
    return 0;
  }
  if ( ex == C_CP )  { 
    nxt = skip_sp(nxt);
    p2 = px_string(nxt, &ln2);
    if ( ln2 > 0 ) {
      *(p2+ln2) = 0;
      nxt = p2 + ln2 + 1;
    } else {
      printf(" ??Command File Name 2:[");
      myputs(nxt);
      printf("] ");
      return 0;
    }
  }
  nxt = skip_sp(nxt);
  if( *nxt != 0x0A ) {
    printf(" ?? ERROR cmd Last:[");
    myputs(nxt);
    printf("] ");
    return 0;
  }
  if( !fatfs_is_ready() && ex != C_MOUNT ) {
    printf(" SD Card Not Ready ?? ");
    return  0;
  }
  switch(ex) {
  case C_CP:
      if( F_copy(p1, p2) == 0 )
        return 0;
      break;
  case C_MKDIR:
      if( f_mkdir(p1) != FR_OK )
         return 0;
      break;
  case C_RM:
  case C_RMDIR:
      if( f_unlink (p1) != FR_OK)
         return 0;
      break;
  case C_TYPE:
      if( F_type(p1) != FR_OK )
         return 0;
      break;
  case C_EDIT:
      tte(ln1 ? p1 : "" );
      break;
  default:
      return 0;
      break;
  }
  ln2 = (int)(nxt - ps);
  return ln2;
}
